@@ -1,11 +1,11 @@
 ;;; osm.el --- OpenStreetMap viewer -*- lexical-binding: t -*-
 
-;; Copyright (C) 2022-2025 Free Software Foundation, Inc.
+;; Copyright (C) 2022-2026 Free Software Foundation, Inc.
 
 ;; Author: Daniel Mendler <mail@daniel-mendler.de>
 ;; Maintainer: Daniel Mendler <mail@daniel-mendler.de>
 ;; Created: 2022
-;; Version: 2.0
+;; Version: 2.1
 ;; Package-Requires: ((emacs "29.1") (compat "30"))
 ;; URL: https://github.com/minad/osm
 ;; Keywords: network, multimedia, hypermedia, mouse
@@ -35,7 +35,7 @@
 ;; locations using regular Emacs bookmarks or create links from Org files
 ;; to locations.  Furthermore the package provides commands to measure
 ;; distances, search for locations and routes by name and to open and
-;; display GPX tracks.
+;; display GPX or TCX tracks.
 
 ;; osm.el requires Emacs 29 and depends on the external `curl' program.
 ;; Emacs must be built with libxml, libjansson, librsvg, libjpeg, libpng
@@ -86,9 +86,8 @@ The server must offer the OSRM API."
 (defcustom osm-server-defaults
   '( :min-zoom 2
      :max-zoom 19
-     :download-batch 8
-     :max-connections 3
-     :subdomains ("a" "b" "c"))
+     :download-batch 5
+     :max-connections 2)
   "Default server properties.
 See also `osm-server-list'."
   :type 'plist)
@@ -115,7 +114,7 @@ See also `osm-server-list'."
       (fr
        :name "Carto(fr)"
        :description "Localized Carto map by OpenStreetMap France"
-       :url "https://%s.tile.openstreetmap.fr/osmfr/%z/%x/%y.png"
+       :url "https://a.tile.openstreetmap.fr/osmfr/%z/%x/%y.png"
        :group "Standard"
        :copyright (,copyright-data
                    "Map style © {OpenStreetMap France|https://www.openstreetmap.fr/mentions-legales/}"
@@ -123,7 +122,7 @@ See also `osm-server-list'."
       (humanitarian
        :name "Humanitarian"
        :description "Humanitarian map provided by OpenStreetMap France"
-       :url "https://%s.tile.openstreetmap.fr/hot/%z/%x/%y.png"
+       :url "https://a.tile.openstreetmap.fr/hot/%z/%x/%y.png"
        :group "Special Purpose"
        :copyright (,copyright-data
                    "Map style © {Humanitarian OpenStreetMap Team|https://www.hotosm.org/}"
@@ -131,7 +130,7 @@ See also `osm-server-list'."
       (cyclosm
        :name "CyclOSM"
        :description "Bicycle-oriented map provided by OpenStreetMap France"
-       :url "https://%s.tile-cyclosm.openstreetmap.fr/cyclosm/%z/%x/%y.png"
+       :url "https://a.tile-cyclosm.openstreetmap.fr/cyclosm/%z/%x/%y.png"
        :group "Transportation"
        :copyright (,copyright-data
                    "Map style © {CyclOSM|https://www.cyclosm.org/} contributors"
@@ -139,7 +138,7 @@ See also `osm-server-list'."
       (openriverboatmap
        :name "OpenRiverBoatMap"
        :description "Waterways map provided by OpenStreetMap France"
-       :url "https://%s.tile.openstreetmap.fr/openriverboatmap/%z/%x/%y.png"
+       :url "https://a.tile.openstreetmap.fr/openriverboatmap/%z/%x/%y.png"
        :group "Transportation"
        :copyright (,copyright-data
                    "Map style © {OpenRiverBoatMap|https://github.com/tilery/OpenRiverboatMap}"
@@ -165,16 +164,14 @@ Allowed keys:
   :max-zoom        Maximum zoom level
   :download-batch  Number of tiles downloaded via a single connection
   :max-connections Maximum number of parallel connections
-  :subdomains      Subdomains used for the %s placeholder
 
-See also `osm-server-defaults' for default values used for a
-server if the property is missing.
+See also `osm-server-defaults' for default values used for a server if
+the property is missing.
 
-The :url of each server should specify %x, %y, %z and %s placeholders
-for the map coordinates.  It can optionally use an %s placeholder
-for the subdomain and a %k placeholder for an apikey.  The apikey
-will be retrieved via `auth-source-search' with the :host set to
-the domain name and the :user to the string \"apikey\"."
+The :url of each server should specify %x, %y and %z placeholders for
+the map coordinates.  It can optionally use a %k placeholder for an
+apikey.  The apikey will be retrieved via `auth-source-search' with the
+:host set to the domain name and the :user to the string \"apikey\"."
   :type '(alist :key-type symbol :value-type plist))
 
 (defcustom osm-copyright t
@@ -408,9 +405,6 @@ Should be at least 7 days according to the server usage policies."
 (defvar osm--download-active nil
   "Globally active download jobs.")
 
-(defvar osm--download-subdomain nil
-  "Subdomain indices to query the servers in a round-robin fashion.")
-
 (defvar-local osm--download-queue nil
   "Buffer-local tile download queue.")
 
@@ -551,7 +545,6 @@ Local per buffer since the overlays depend on the zoom level.")
 (defun osm--tile-url (x y zoom)
   "Return tile url for coordinate X, Y and ZOOM."
   (let ((url (osm--server-property :url))
-        (sub (osm--server-property :subdomains))
         (key (osm--server-property :key)))
     (when (and (string-search "%k" url) (not key))
       (require 'auth-source)
@@ -570,10 +563,7 @@ Local per buffer since the overlays depend on the zoom level.")
         (setf (plist-get (alist-get osm-server osm-server-list) :key) key)))
     (format-spec
      url `((?z . ,zoom) (?x . ,x) (?y . ,y)
-           (?k . ,(if (functionp key) (funcall key) key))
-           (?s . ,(nth (mod (alist-get osm-server osm--download-subdomain 0)
-                            (length sub))
-                       sub))))))
+           (?k . ,(if (functionp key) (funcall key) key))))))
 
 (defun osm--tile-file (x y zoom)
   "Return tile file name for coordinate X, Y and ZOOM."
@@ -609,8 +599,8 @@ Local per buffer since the overlays depend on the zoom level.")
       (when (equal status "200")
         (ignore-errors (rename-file file (string-remove-suffix ".tmp" file) t))
         (osm--each
-          (when (and (= osm--zoom zoom) (eq osm-server server))
-            (osm--display-tile x y (osm--get-tile x y)))))
+         (when (and (= osm--zoom zoom) (eq osm-server server))
+           (osm--display-tile x y (osm--get-tile x y)))))
       (cl-callf2 delete (list server zoom x y) osm--download-active)
       (delete-file file)))
   output)
@@ -619,8 +609,7 @@ Local per buffer since the overlays depend on the zoom level.")
   "Build download command."
   (let* ((count 0)
          (batch (osm--server-property :download-batch))
-         (subs (length (osm--server-property :subdomains)))
-         (parallel (* subs (osm--server-property :max-connections)))
+         (parallel (osm--server-property :max-connections))
          args jobs job)
     (while (and (< count batch)
                 (setq job (nth (* count parallel) osm--download-queue)))
@@ -633,10 +622,8 @@ Local per buffer since the overlays depend on the zoom level.")
         (push job osm--download-active)
         (cl-incf count)))
     (osm--each
-      (dolist (job jobs)
-        (cl-callf2 delq job osm--download-queue)))
-    (cl-callf (lambda (s) (mod (1+ s) subs))
-        (alist-get osm-server osm--download-subdomain 0))
+     (dolist (job jobs)
+       (cl-callf2 delq job osm--download-queue)))
     (cons `("curl" "--disable" "--write-out" "%{http_code} %{filename_effective}\n"
             ,@(split-string-and-unquote osm-curl-options) ,@(nreverse args))
           jobs)))
@@ -644,8 +631,7 @@ Local per buffer since the overlays depend on the zoom level.")
 (defun osm--download ()
   "Download next tiles from the queue."
   (when (and (< (length (alist-get osm-server osm--download-processes))
-                (* (length (osm--server-property :subdomains))
-                   (osm--server-property :max-connections)))
+                (osm--server-property :max-connections))
              osm--download-queue)
     (pcase-let ((`(,command . ,jobs) (osm--download-command))
                 (dir (file-name-concat (expand-file-name osm-tile-directory)
@@ -671,8 +657,8 @@ Local per buffer since the overlays depend on the zoom level.")
           (force-mode-line-update t)
           (osm--download)))
        (alist-get server osm--download-processes))
-       (force-mode-line-update t)
-       (osm--download))))
+      (force-mode-line-update t)
+      (osm--download))))
 
 (defun osm-mouse-drag (event)
   "Handle drag EVENT."
@@ -982,13 +968,13 @@ Local per buffer since the overlays depend on the zoom level.")
            (pcase-let ((`(,lat ,lon ,zoom) (bookmark-prop-get bm 'coordinates)))
              (funcall fun 'osm-bookmark lat lon zoom (car bm))))
   (cl-loop for (dname id segs waypoints) in osm--datasets do
-    (when-let ((start (caar segs)))
-      (funcall fun id (car start) (cdr start) 10
-               (propertize dname 'osm-dataset dname)))
-    (cl-loop for (lat lon name) in waypoints do
-             (funcall fun id lat lon 15
-                      (propertize (format "%s [%s]" name dname)
-                                  'osm-dataset dname))))
+           (when-let* ((start (caar segs)))
+             (funcall fun id (car start) (cdr start) 10
+                      (propertize dname 'osm-dataset dname)))
+           (cl-loop for (lat lon name) in waypoints do
+                    (funcall fun id lat lon 15
+                             (propertize (format "%s [%s]" name dname)
+                                         'osm-dataset dname))))
   (cl-loop for (lat lon name) in osm--track do
            (funcall fun 'osm-track lat lon 15 name)))
 
@@ -1202,8 +1188,8 @@ xmlns='http://www.w3.org/2000/svg' xmlns:xlink='http://www.w3.org/1999/xlink'>
   (clear-image-cache t) ;; Make absolutely sure that the tiles are redrawn.
   (setq osm--tile-cache nil)
   (osm--each
-    (setq osm--overlays nil)
-    (osm--update)))
+   (setq osm--overlays nil)
+   (osm--update)))
 
 (defun osm--header-button (text action)
   "Format header line button with TEXT and ACTION."
@@ -1607,7 +1593,7 @@ When called interactively, call the function `osm-home'."
 
 (defun osm--track-delete ()
   "Delete track way point."
-  (when-let ((idx (osm--track-index)))
+  (when-let* ((idx (osm--track-index)))
     ;; Delete pin
     (cl-callf2 delq (nth idx osm--track) osm--track)
     (setq osm--pin nil
@@ -1626,9 +1612,9 @@ When called interactively, call the function `osm-home'."
 
 (defun osm--track-rename ()
   "Rename track way point."
-  (when-let ((pt (nth (osm--track-index) osm--track))
-             (old-name (caddr pt))
-             (new-name (read-from-minibuffer "New name: " old-name nil nil nil old-name)))
+  (when-let* ((pt (nth (osm--track-index) osm--track))
+              (old-name (caddr pt))
+              (new-name (read-from-minibuffer "New name: " old-name nil nil nil old-name)))
     (setf (caddr pt) new-name
           (cadddr osm--pin) new-name)
     (osm--revert)))
@@ -1818,37 +1804,82 @@ See `osm-search-server' and `osm-search-language' for customization."
 
 ;;;###autoload
 (defun osm-open (file)
-  "Show the tracks of GPX FILE in an `osm-mode' buffer."
-  (interactive "fGPX file: ")
+  "Show the tracks of GPX or TCX FILE in an `osm-mode' buffer."
+  (interactive "fGPX or TCX file: ")
   (osm--check-libraries)
   (let ((dom (with-temp-buffer
                (insert-file-contents file)
-               (libxml-parse-xml-region (point-min) (point-max)))))
-    (unless (eq 'gpx (dom-tag dom))
-      (setq dom (dom-child-by-tag dom 'gpx)))
-    (unless (and dom (eq 'gpx (dom-tag dom)))
-      (error "Not a GPX file"))
-    (osm--add-dataset
-     (abbreviate-file-name file)
-     'osm-file
+               (libxml-parse-xml-region (point-min) (point-max))))
+        (file (abbreviate-file-name file)))
+    (if (eq 'TrainingCenterDatabase (dom-tag dom))
+        (osm--add-dataset file 'osm-file
+                          (osm--tcx-track dom)
+                          (osm--tcx-waypoints dom))
+      (unless (eq 'gpx (dom-tag dom))
+        (setq dom (dom-child-by-tag dom 'gpx)))
+      (unless (and dom (eq 'gpx (dom-tag dom)))
+        (error "Not a GPX or TCX file"))
+      (osm--add-dataset file 'osm-file
+                        (osm--gpx-track dom)
+                        (osm--gpx-waypoints dom)))))
+
+;; Prevent deprecation warning
+(defalias 'osm--dom-text 'dom-text)
+
+(defsubst osm--tcx-position (pos)
+  "Convert POS to a cons of (lat . lon)."
+  (let ((lat (dom-child-by-tag pos 'LatitudeDegrees))
+        (lon (dom-child-by-tag pos 'LongitudeDegrees)))
+    (cons (string-to-number (osm--dom-text lat))
+          (string-to-number (osm--dom-text lon)))))
+
+(defun osm--tcx-waypoints (dom)
+  "Return waypoints contained in tcx DOM."
+  (cl-loop
+   with courses = (dom-child-by-tag dom 'Courses)
+   for course in (dom-by-tag courses 'Course) nconc
+   (cl-loop
+    for pt in (dom-by-tag course 'CoursePoint)
+    for pos = (osm--tcx-position (dom-child-by-tag pt 'Position))
+    for name = (dom-by-tag pt 'Name)
+    collect (list (car pos) (cdr pos) (osm--dom-text name)))))
+
+(defun osm--tcx-track (dom)
+  "Return track points contained in tcx DOM."
+  (cl-loop
+   with activities = (dom-child-by-tag dom 'Activities)
+   for activity in (dom-by-tag activities 'Activity) nconc
+   (cl-loop
+    for lap in (dom-by-tag activity 'Lap) nconc
+    (cl-loop
+     for track in (dom-by-tag lap 'Track) collect
      (cl-loop
-      for trk in (dom-children dom)
-      if (eq (dom-tag trk) 'trk) nconc
-      (cl-loop
-       for seg in (dom-children trk)
-       if (eq (dom-tag seg) 'trkseg) collect
-       (cl-loop
-        for pt in (dom-children seg)
-        if (eq (dom-tag pt) 'trkpt) collect
-        (cons (string-to-number (dom-attr pt 'lat))
-              (string-to-number (dom-attr pt 'lon))))))
-     (cl-loop
-      for pt in (dom-children dom)
-      if (eq (dom-tag pt) 'wpt) collect
-      (list (string-to-number (dom-attr pt 'lat))
-            (string-to-number (dom-attr pt 'lon))
-            (with-no-warnings
-              (dom-text (dom-child-by-tag pt 'name))))))))
+      for pt in (dom-by-tag track 'Trackpoint)
+      for pos = (dom-child-by-tag pt 'Position)
+      if pos collect (osm--tcx-position pos))))))
+
+(defun osm--gpx-track (dom)
+  "Return track points contained in gpx DOM."
+  (cl-loop
+   for trk in (dom-children dom)
+   if (eq (dom-tag trk) 'trk) nconc
+   (cl-loop
+    for seg in (dom-children trk)
+    if (eq (dom-tag seg) 'trkseg) collect
+    (cl-loop
+     for pt in (dom-children seg)
+     if (eq (dom-tag pt) 'trkpt) collect
+     (cons (string-to-number (dom-attr pt 'lat))
+           (string-to-number (dom-attr pt 'lon)))))))
+
+(defun osm--gpx-waypoints (dom)
+  "Return waypoints contained in gpx DOM."
+  (cl-loop
+   for pt in (dom-children dom)
+   if (eq (dom-tag pt) 'wpt) collect
+   (list (string-to-number (dom-attr pt 'lat))
+         (string-to-number (dom-attr pt 'lon))
+         (osm--dom-text (dom-child-by-tag pt 'name)))))
 
 (defun osm--add-dataset (name id track waypoints)
   "Add dataset with NAME and ID consisting of TRACK and WAYPOINTS."
@@ -1870,12 +1901,12 @@ See `osm-search-server' and `osm-search-language' for customization."
 
 (defun osm--server-annotation (cand)
   "Annotation for server CAND."
-  (when-let ((copyright (osm--server-property :copyright (get-text-property 0 'osm--server cand)))
-             (str
-              (replace-regexp-in-string
-               "{\\(.*?\\)|.*?}"
-               (lambda (str) (match-string 1 str))
-               (string-join (ensure-list copyright) " | ") copyright)))
+  (when-let* ((copyright (osm--server-property :copyright (get-text-property 0 'osm--server cand)))
+              (str
+               (replace-regexp-in-string
+                "{\\(.*?\\)|.*?}"
+                (lambda (str) (match-string 1 str))
+                (string-join (ensure-list copyright) " | ") copyright)))
     (concat (propertize " " 'display `(space :align-to (- right ,(length str) 2)))
             " "
             str)))
@@ -1944,13 +1975,13 @@ If prefix ARG is given, store URL as Elisp expression."
 (cl-defun osm-add-server (server
                           &rest properties
                           &key name description group url ext max-connections
-                          max-zoom min-zoom download-batch subdomains copyright)
+                          max-zoom min-zoom download-batch copyright)
   "Add SERVER with PROPERTIES to `osm-server-list'.
 The properties are checked as keyword arguments.  See
 `osm-server-list' for documentation of the keywords."
   (declare (indent 1))
   (ignore name description group url ext max-connections max-zoom
-          min-zoom download-batch subdomains copyright)
+          min-zoom download-batch copyright)
   (dolist (sym '(:name :description :group :url))
     (unless (stringp (plist-get properties sym))
       (error "Server property %s is required" sym)))
